# Dieses Programm stellt zwei Funktionen zur Verfgung, die einen +String+ in einer while-Sprache in seine
# Bestandteile zerlegen bzw. ihren Tokenklassen zuordnen. Diese Funktionen werden an folgendem Beispiel getestet:
#
#	d:=80;
#	while (x>=y) do
#	  d:=(d div 2);
#	  x:=(x-y);
#	od
#
# Author:: Lars Gerhard

# Diese Dateien werden fr die Unit-Tests bentigt.
require 'runit/testcase'
require 'runit/cui/testrunner'
require 'runit/testsuite'

# ==zerlege(whtxt)
# Diese Methode zerlegt einen Eingabestring in seine Bestandteile, die durch Ruby-Patterns gegeben sind
# Allerdings werden die einzelnen Teile nicht den Tokenklassen zugeordnet, sondern nur in ein Array ge
# schrieben, welches von der Methode dann zurckgegeben wird. Die Patterns sind:
#
# * name 				= [a-zA-Z][a-zA-Z0-9]*
# * zahl 				= [1-9]+[0-9]*|0
# * addOp 			= +|-
# * mulOp 			= *|div 
# * auf 				= (
# * zu          = )
# * ergibt      = :=
# * strichpunkt = ;

def zerlege(whtxt)
	comp=whtxt.scan(/;|:=|\+|\-|div|mod|\(|\)|[a-zA-Z][a-zA-Z0-9]*|[1-9]+[0-9]*|0/)
	return comp
end

# ==bestimme(whtxt)
# Die Methode "bestimme" scant einen String von Anfang bis Ende und ordnet sie den Tokenklassen zu.
# Diese Klassen sind in einem Array und einem Hash-Set als regulre Ausdrcke gespeichert. Die Ordnung 
# des Token-Array ergibt gleichzeitig auch die Prioritt der Tokens. Hier also in absteigender Reihenfolge: 
# +name+, +zahl+, +addOp+, +mulOp+, +auf+, +zu+, +ergibt+, +strichpunkt+ (d.h. +strichpunkt+ hat die niedrigste 
# Prioritt). Ob diese Anordnung sinnvoll ist, spielt fr den Scanner keine Rolle.
#
# Das Hash-Set wird nur verwendet, weil der Programmtext leichter verstndlich ist, als bei einem mehr
# dimensionalen Array.
#
# Die Zuordnung zu den einzelnen Tokens: Dazu wird die Eingabe zunchst zerlegt. Danach werden die Tokens in der 
# entsprechend durch Prioroitten festgelegten Reihenfolge auf die Bestandteile angewendet. Sobald ein Treffer 
# gefunden ist, wird +accumulated+ auf +true+ gesetzt und der nchste Bestandteil der Eingabe geprft.

def bestimme(whtxt)
	#Definition der einzelnen Tokens als regulre Audrcke
	name = Regexp.new('[a-zA-Z][a-zA-Z0-9]*')
	zahl = Regexp.new('[1-9]+[0-9]*|0')
	addOp = Regexp.new('\+|\-')
	mulOp = Regexp.new('\*|div|mod')
	auf = Regexp.new('\(')
	zu = Regexp.new('\)')
	ergibt = Regexp.new(':=')
	strichpunkt = Regexp.new(';')

	tokenArray=[name,zahl,addOp,mulOp,auf,zu,ergibt,strichpunkt]
	
	tokens=Hash.new
	tokens[name]="name"
	tokens[zahl]="zahl"
	tokens[addOp]="addOp"
	tokens[mulOp]="mulOp"
	tokens[auf]="auf"
	tokens[zu]="zu"
	tokens[ergibt]="ergibt"
	tokens[strichpunkt]="strichpunkt"

	zerlege(whtxt).each do |comp|	
		accumulated=false
		tokenArray.each do |regex|
			if accumulated == false
				if not (regex=~comp)==nil
					print("- ",tokens[regex], " << ",comp," >>\n")
					accumulated=true
				end
			end
		end		
	end
	
end

example="d:=0;\nwhile x>=y do\nd:=(d div 123);\nx:=(x-y); od"

print("+------------------------------------------------+\n")
print("| Zerlegung ohne Zuordnung zu den Token-Klassen: |\n")
print("+------------------------------------------------+\n\n")
zerlege(example).each do |comp| 
	print("<<",comp,">>  ")
end

print("\n\n")
print("+-----------------------------------------------+\n")
print("| Zerlegung mit Zuordnung zu den Token-Klassen: |\n")
print("+-----------------------------------------------+\n\n")
print("Akkumulierte Tokens: \n--------------------\n\n")
bestimme(example)

# Testklasse zur Zerlegung eines While-Ausdrucks in seine Tokens
class Testing_class < RUNIT::TestCase

  # == test_zerlege
	# Der folgende While-Ausdruck soll korrekt zerlegt werden:
	# "d:=0;\nwhile x>=y do\nd:=(d div 123);\nx:=(x-y); od"
	def test_zerlege
	string = "d:=0;\nwhile x>=y do\nd:=(d div 123);\nx:=(x-y); od"
	comp = ["d",":=","0",";","while","x","y","do","d",":=","(","d","div","123",")",";","x",":=","(","x","-","y",")",";","od"]
	assert_equal(comp,zerlege(string))
  end
end
