# Dieses Programm berechnet Markov-Algorithmen. Dazu werden zwei Input-Dateien abgefragt von denen die erste mit der
# Endung *.alg den Algorithmus enthlt (mit einer Zeile pro Regel und ohne berflssige Blanks) und die zweite mit der
# Endung *.inp den Text enthlt den der Markov-Interpreter bearbeiten soll (eine Zeile, ohne Return oder Blanks). Beide 
# Dateinamen werden ohne Endung eingegeben. 
# Das Programm generiert dann bei erfolgreichem Ablauf folgende Dateien (+<name>+ steht fr den Namen der Algorithmus-Datei):
# * <name>_comp.rb  : Ein Ruby-Programm, das den Compiler fr den Markov-Algorithmus darstellt.
# * <name>_tracer.rb: Ein eigentlich nur temporres Programm, welches zur Erstellung der Trace-Datei dient.
# * <name>_trace.txt: Eine Textdatei, die die Trace, also die Ergebnisse der einzelnen Regelanwendungen, enthlt.
# Desweiteren wird der Algorithmus auf die Eingabedatei angewandt und das Ergebnis auf dem Bildschirm (nach dem Algorithmus
# und dem Ruby-Compile-Quellcode) ausgegeben.
#
# ===Beispiel:
# 1. Algorithmusdatei test.alg:
#		 I*->H*
#		 IH->HH
#		 HD->DEH
#		 ED->DE
#		 D->
#		 *I->D*
#		 H->
#		 *->
#		 E->I
# 2. Inputdatei test.inp:
#		 III*II
# 3. Ergebnis der Anwendung auf den Input:
#		 IIIIII
# 4. Generierte Trace-Datei test_trace.txt:
#		 III*II
#		 IIH*II
#		 IHH*II
#		 HHH*II
#		 HHHD*I
#		 HHDEH*I
#		 HDEHEH*I
#		 DEHEHEH*I
#		 EHEHEH*I
#		 EHEHEHD*
#		 EHEHEDEH*
#		 EHEHDEEH*
#		 EHEDEHEEH*
#		 EHDEEHEEH*
#		 EDEHEEHEEH*
#		 DEEHEEHEEH*
#		 EEHEEHEEH*
#		 EEEEHEEH*
#		 EEEEEEH*
#		 EEEEEE*
#		 EEEEEE
#		 IEEEEE
#		 IIEEEE
#		 IIIEEE
#		 IIIIEE
#		 IIIIIE
#		 IIIIII
#
# Author:: Lars Gerhard, Sebastian Haffner

# Diese Dateien werden fr die Unit-Tests bentigt.
require 'runit/testcase'
require 'runit/cui/testrunner'
require 'runit/testsuite'

# == markovgen(input)
# Diese Methode ist der Hauptbestandteil des Markov-CompilerCompilers. Sie erhlt den Markov-Algorithmus in Textform als
# Parameter und liefert den Compiler als Ruby-Programm ebenfalls in Textform zurck. Im Wesentlichen werden folgende Bausteine
# zusammengestellt:
#
# 1. Methodenheader:
#	   def markovc(input)
# 2. If-Anweisungen fr jede Einzelne Markov-Regel:
# 	 if not (input=~/linkeSeite/)==nil 
#			   markovc(input.sub!(/linkeSeite/, "rechteSeite")
#	   end
# 3. Ergebnisrckgabe und Methodenende:
# 	 return input
# 	 end
#
# Im regulren Ausdruck 'linkeSeite' werden die Metazeichen von Ruby (?.|(){+\^$*[]) escapet, also mit vorangestelltem Backslash
# in den Ausdruck bernommen
def markovgen(input)
	result=("def markovc(input)\n\n")
	
	#Aufteilen der Eingabe in einzelne Regeln (durch einzelne Zeilen erkennbar)
	input.split(/[\n]/).each do |rule|
	
		#Aufteilen der Regeln in linke und rechte Seiten am "->"
		leftside=rule.split("->")[0]
		rightside=rule.split("->")[1]
		
		#Escapen der Metazeichen von Ruby in der linken Seite der Markov-Regeln. Auf
		#der rechten Seite ist dies nicht ntig, da diese in der If-Anweisung nur als
		#String genutzt wird.
		leftside.gsub!(/[?.|(){+\^$*\[\]]/, '\\\\\&')
		
		#Zusammensetzen der If-Anweisung
		result=result + "  #Regel: "+ rule +"\n"
		result=result + "  if not (input=~/"+ leftside +"/)==nil\n"
		if rightside==nil
			result=result + "    markovc(input.sub!(/"+leftside+"/, ''))\n"
		else
			result=result + "    markovc(input.sub!(/"+leftside+"/, '"+rightside+"'))\n"
		end
		result=result + "  end\n\n"
	end
	
	result=result+"  return input\nend\n"
	return result
end

#Abfrage des Dateinamens fr den Algorithmus
print("Dateiname der Definition des Markov-Algorithmus: (*.alg) \n\n # ")
fname = readline
fname.gsub!(/[\n]/,'')

#Abfrage des Dateinamens fr den Algorithmus
print("\nDateiname des Markov-Input: (*.inp) \n\n # ")
fname2 = readline
fname2.gsub!(/[\n]/,'')

#Abfrage des Dateinamens mit dem zu interpretierenden Text
inputfile = File.open(fname+".alg","r")
intext = inputfile.read
inputfile.close

#Anzeigen des Algorithmus
print("\n++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
print("Markov-Algorithmus: \n")
print("------------------- \n")
print(intext,"\n\n\n")

#Generieren des Markov-Compiler (Ruby-Interpreter)
outputfile = File.new(fname+"_comp.rb",  "w+")
outputfile.print(markovgen(intext))
outputfile.close

#Anzeigen des Markov-Compiler Quelltext
print("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
print("Ruby-Interpreter: \n")
print("----------------- \n")
print(markovgen(intext))

#Auslesen des zu interpretierenden Textes
runfile = File.open(fname2+".inp","r")
runtext = runfile.read
runfile.close
require fname+"_comp.rb"

#Anzeigen der Anwendung des Compilers auf den Input
print("\n\n++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
print("Anwendung des Interpreters auf den Input: \n")
print("----------------------------------------- \n")
print(markovc(runtext))

#Der Trace-Generator:
#Der Trace-Generator entsteht aus dem Markov-Generator (Ruby-Interpreter) durch Ersetzen des
#Methodennamens und Entfernen der Rekursion.
tracefileGen = File.new(fname+"_tracer.rb","w+")
tracefileGenText = markovgen(intext).gsub!(/def markovc/, "def tracegen")
tracefileGenText = tracefileGenText.gsub!(/markovc/, "return")
tracefileGen.print tracefileGenText
tracefileGen.close

#Erneutes Auslesen des zu interpretierenden Text
runfile = File.open(fname2+".inp","r")
tracesrc = runfile.read
runfile.close

require fname+"_tracer.rb"
tracelog = File.new(fname+"_trace.txt", "w+")

actline = tracesrc.dup
nextline = actline.dup
nextline = tracegen(nextline)

#Der Algorithmus ist beendet, wenn keine Ersetzung mehr erfolgt, also actline=nextline ist
while not (actline==nextline) do
	tracelog.print(actline,"\n")
	actline=nextline.dup
	nextline=tracegen(nextline)
end
tracelog.print(nextline,"\n") #Die letzte Zeile
tracelog.close

# Testklasse
class Testing_class < RUNIT::TestCase
	
	#== test_markovgen
	# Diese Methode testet, ob der Ruby-Code korrekt generiert wird.
	def test_markovgen
	inputfile = File.open("test.alg","r")
	alg = inputfile.read
	inputfile.close
	inputfile = File.open("test.rb","r")
	algToRb = inputfile.read
	inputfile.close
	assert_equal(algToRb,markovgen(alg))
  end
	
	# Diese Methode testet, ob der Ruby-Code korrekt angewendet wird.
  def test_markovc
	inputfile = File.open("test.inp","r")
	inp = inputfile.read
	inputfile.close
	erg="IIIIII\n"
	assert_equal(erg,markovc(inp))
  end
end

