# Author::    Jens Bachmann  (mailto:jens.bachmann@unibw.de)
# Author::		Daniel Borkowitz (mailto:daniel.borkowitz@unibw.de)
# Author::		Stefan Mayer (mailto:stefan.mayer@unibw.de)
# Author::		Roman Krhling (mailto:roman.kraehling@unibw.de)
# Copyright:: Copyright (c) 2006 The Heimschlfers, UniBw Mnchen

# Klasse Menge dient zur Modellierung von Mengen, hierbei werden die Elemente der Menge
# in einem Ruby-Array (@array) gehalten
class Menge

# Erschafft ein Mengenobjekt aus einem Arrayparameter, entfernt 
# Doubletten und sortiert die Menge
	def initialize(array)
		@array = array
		@array.uniq!()
		@array.sort!()
	end
	
# Konvertiert die Menge in String 
# Bsp.: [1,2,3,4,5] -> "{ 1,2,3,4,5 }"
	def to_s
		return "{}" if @array == nil
		n=0
		str = "{ " + @array[0].to_s
		while n<=@array.length
			n+=1
			str += ',' + @array[n].to_s if @array[n]!=nil
		end
		return str + " }"
	end
	
# Fgt ein Element an die Menge an und sortiert die Menge und entfernt Doubletten
	def add(elem)
		@array+=[elem]
		@array.uniq!()
		@array.sort!()
	end
	
# Prft ob die Menge das Element enthlt
	def enthaelt?(x)
		@array.include?(x)
	end

# Getter fr @array	
	def array()
		@array
	end

# Liefert die Anzahl der Elemente der Menge
	def length()
		@array.length()
	end

# Vereinigung zweier Mengen, dabei wird berprft ob die Menge (Parameter) nil ist	
	def +(menge)
		if menge!=nil then Menge.new(@array + menge.array()) else return Menge.new(@array) end
	end

# Differenz zweier Mengen, dabei wird berprft ob die Menge (Parameter) nil ist	
	def -(menge)
		if menge!=nil then Menge.new(@array - menge.array()) else return Menge.new(@array) end
	end

# Prft Gleichheit zweier Mengen, dabei wird berprft ob die Menge (Parameter) nil ist	
	def ==(menge)
		if menge!=nil then (@array==menge.array()) else return false end
	end

# Ist Iterator ber die Elemente der Menge
	def each &b
		@array.each &b
	end

# Liefert eine wahre Kopie der Menge	
	def copy()
		temp=Menge.new([])
		each{|i| temp.add(i)}
		return temp
	end

# Speichert das Objekt Menge auf der Harddisk mit Dateiname als Parameter mittels Marshall	
	def speichern(dateiname)
		File.open(dateiname, "w") do |f|
			Marshal.dump(self,f)
		end
	end

# Ldt das Objekt Menge von der Harddisk mit Dateiname als Parameter mittels Marshall	
	def laden(dateiname)
		File.open(dateiname, "r") do |f|
			@array+=Marshal.load(f).array()
		end
	end
end


#######

# Klasse Relation dient zur Modellierung von Relationen, hierbei wird die Basismenge im 
# Attribut @basis als Menge gehalten. Zur Basis gibt es eine Vorgnger- und eine 
# Nachfolgeradjazenzliste (@vor und @nach). Diese sind als Ruby-Hashs von Mengen abgelegt,
# der Schlssel ist jeweils das Element aus der Basis.
class Relation

# Konstruktor der Klasse Relation erhlt ein Array von Paaren (ein Paar ist ein Array mit zwei Elementen).
# Fr @vor und @nach werden neue Hashs angelegt. Die Basis wird anhand des bergegebenen Parameters (rel) berechnet.
# Die Vorgnger und Nachfolger werden nun anhand der Basis und rel berechnet.
	def initialize(rel)
		@vor = Hash.new()
		@nach = Hash.new()
		@basis=Menge.new([]) 
		if rel!=[[]]
		@basis=Menge.new([])
		rel.each{|i| @basis.add(i[0]);@basis.add(i[1])}
	  each {|i| @vor[i]=berechneVorgaenger(i,rel)}
	  each {|i| @nach[i]=berechneNachfolger(i,rel)}
		end
	end

private
# Berechnet die Vorgnger eines Elements und gegebener Relation.
	def berechneVorgaenger(elem,rel)
		temp= Menge.new([])
		rel.each{|i| temp.add(i[0]) if i[1]==elem}
		return temp
	end
	
# Berechnet die Nachfolger eines Elements und gegebener Relation.
	def berechneNachfolger(elem,rel)
		temp= Menge.new([])
		rel.each{|i| temp.add(i[1]) if i[0]==elem}
		return temp
	end
	
public
# Getter fr die Vorgngeradjazenzliste.
	def vor 
		@vor
	end

# Setter fr die Vorgngeradjazenzliste.
	def vor!(vor)
		@vor=vor
	end
	
# Getter fr die Nachfolgeradjazenzliste.
	def nach
		@nach
	end

# Setter fr die Nachfolgeradjazenzliste.
	def nach!(nach)
		@nach=nach
	end
	
# Getter fr die Basismenge.
	def basis
		@basis
	end
	
# Setter fr die Basismenge.
	def basis!(basis)
		@basis = basis
	end

# Konvertiert die Relation in einen String.
# 
# Bsp.: [[1,2],[2,3],[3,4],[2,5]]
#
# "Basismenge:
#    { 1,2,3,4,5 }
# Adjazenzliste (Vorgaenger):
#   5{ 2 }1{  }2{ 1 }3{ 2 }4{ 3 }
# Adjazenzliste (Nachfolger):
#   5{  }1{ 2 }2{ 3,5 }3{ 4 }4{  }"
	def to_s
			return "{}" if @basis == nil
			str = "Basismenge: \n   " + @basis.to_s + "\n" +
			"Adjazenzliste (Vorgaenger): \n   " + @vor.to_s + "\n"+
			"Adjazenzliste (Nachfolger): \n   " + @nach.to_s + "\n \n"
			return str
	end

# Speichert das Objekt Relation auf der Harddisk mit Dateiname als Parameter mittels Marshall	
	def speichern(dateiname)
		File.open(dateiname, "w") do |f|
			Marshal.dump(@basis,f)
			Marshal.dump(@vor,f)
			Marshal.dump(@nach,f)
		end
	end

# Ldt das Objekt Relation von der Harddisk mit Dateiname als Parameter mittels Marshall	
	def laden(dateiname)
		File.open(dateiname, "r") do |f|
			@basis=Marshal.load(f)
			@vor=Marshal.load(f)
			@nach=Marshal.load(f)
		end
	end
	
# Fgt ein Tupel (x,y) zur Relation hinzu. Hierbei wird die Relation verndert.
	def addPaar!(x,y)
		@basis.add(x)
		@basis.add(y)
		@nach[x] = Menge.new([y]) + @nach[x]
		@vor[y] = Menge.new([x]) + @vor[y]
	end

# Prft ob die Relation das Tupel (x,y) enthlt.
	def enthaeltPaar?(x,y)
		return (@nach[x].enthaelt?(y) && @vor[y].enthaelt?(x))
	end

# Liefert die Menge der Vorgnger fr das Element x aus der Basismenge.
	def vorgaenger(x)
		if @vor[x]!=nil then return @vor[x] else return Menge.new([]) end
	end

# Liefert die Menge der Nachfolger fr das Element x aus der Basismenge.
	def nachfolger(x)
		if @nach[x]!=nil then return @nach[x] else return Menge.new([]) end
	end

# Prft Gleichheit zweier Relationen, dabei wird berprft ob die Relation (Parameter) nil ist.	
	def ==(rel)
		if rel!=nil then return ((@basis==rel.basis) && (@vor==rel.vor) && (@nach==rel.nach)) else
		return false end
	end
	
# Iterator ber die Elemente der Basismenge der Relation.
	def each &b
		@basis.each &b
	end

# Vereinigung zweier Relationen, dabei werden die Basen zusammengefasst und danach ber die entstandene
# Basis iteriert. Dabei werden die Vorgnger und Nachfolger aus beiden Relationen neu berechnet.
  def vereinigt(rel)
    erg = Relation.new([[]])
    erg.basis!(@basis + rel.basis)
    erg.each{|i| erg.nach[i] = (@nach[i]=nil ? Menge.new([]) : @nach[i]) + rel.nach[i] }
    erg.each{|i| erg.vor[i]  = (@vor[i] =nil ? Menge.new([]) : @vor[i])  + rel.vor[i] }
    return erg
  end

# Komposition zweier Relationen, dabei werden die Basen zusammengefasst und danach ber die entstandene
# Basis iteriert. Dabei werden die Vorgnger und Nachfolger aus beiden Relationen neu berechnet.
	def komp(rel)
		erg=Relation.new([[]])
		erg.basis!(@basis+rel.basis)
			erg.each{|i| erg.nach[i]=Menge.new([]); erg.vor[i]=Menge.new([])}
			erg.each{|i| @nach[i].each{|j| erg.nach[i]+=rel.nachfolger(j)}}
			erg.each{|i| rel.vor[i].each{|j| erg.vor[i]+=vorgaenger(j)}}
			return erg
	end

# Liefert eine wahre Kopie der Relation und ihrer Attribute.
	def copy()
		temp = Relation.new([[]])
		@basis.each{|i| temp.basis.add(i)}
		@vor.each_key{|i| temp.vor[i]=@vor[i].copy}
		@nach.each_key{|i| temp.nach[i]=@nach[i].copy}
		return temp
	end

# Berechnet die reflexive Hlle der Relation.
	def reflexiv
		temp=copy
		temp.each{|i| temp.addPaar!(i,i)}
		return temp
	end

# Berechnet die transitive Hlle der Relation mittels Warshall-Algorithmus.
	def transitiv
		temp=copy
		temp.basis.length().times do
		  temp.each{|i| temp.nachfolger(i).each{|j| temp.nach[i]+=temp.nachfolger(j)}}
		  temp.each{|i| temp.nachfolger(i).each{|j| temp.vor[j]+=Menge.new([i])}}
		end
		return temp
  end
end


# testing
####################################
#tmprel1= [[1,1],[1,12],[2,13],[4,13]]
#rel1 = Relation.new(tmprel1)
#tmprel2= [[12,400],[13,500],[13,600],[12,100],[14,600]]
#rel2 = Relation.new(tmprel2)
tmprel3= [[1,2],[2,3],[3,4],[2,5]]
rel3 = Relation.new(tmprel3)
#puts rel3.basis.enthaelt
puts rel3.transitiv()
puts "Alt:",rel3
#puts rel1==rel1
#puts rel1==rel2
#puts rel1
#rel1.addPaar!(1,5)
#puts rel1
#puts rel1.enthaeltPaar?(1,5)
#puts rel1.enthaeltPaar?(2,5)
#puts rel1.vorgaenger(13)
#puts rel1.vorgaenger(2)
#puts rel1.nachfolger(13)
#puts rel1.nachfolger(1)
#puts rel1.vereinigt(rel2)
#puts rel1
#puts rel1.vereinigt(rel1)

#puts rel1.komp(rel1)
=begin
		temp.each do |i| temp_m=temp.nachfolger(i)-Menge.new([i])
			temp.basis!(temp.basis-Menge.new([i]))
			temp.basis.each do |j| temp.nach[j]=(temp.nachfolger(j)+temp_m) if temp.nachfolger(j).enthaelt?(i) 
			end
		end
=end

=begin
# Vereinigung zweier Relationen, dabei werden die Basen zusammengefasst und danach ber die entstandene
# Basis iteriert. Dabei werden die Vorgnger und Nachfolger aus beiden Relationen neu berechnet.
  def vereinigt(rel)
    erg = Relation.new([[]])
    erg.basis!(@basis + rel.basis)
    erg.each{|i| if @nach[i]!=nil then erg.nach[i] = @nach[i] + rel.nach[i] else erg.nach[i] = Menge.new([]) + rel.nach[i] end}
    erg.each{|i| if @vor[i]!=nil then erg.vor[i] = @vor[i] + rel.vor[i] else erg.vor[i] = Menge.new([]) + rel.vor[i] end}
    return erg
  end
=end